'''
These functions prep the data for trend analyses

'''

import numpy as np
import pandas as pd

def _define_seasons(tdf,
                    seasons_per_year,
                    trend_end_month):
    
    # Copy the DataFrame
    df = tdf.data.copy()
    
    # Add columns for the data frequency and number of seasons per year
    freq_dict = {
        1:'Annually',
        2:'Semiannually',
        3:'Triannually',
        4:'Quarterly',
        6:'Bimonthly',
        12:'Monthly'
        }
    df[tdf.freq_col] = freq_dict[seasons_per_year]
    df[tdf.freq_num_col] = seasons_per_year
    
    # Determine trend year (where year is aligned with trend period)
    df[tdf.trend_year_col] = np.where(df[tdf.month_col] > trend_end_month,
                                      df[tdf.year_col] + 1,
                                      df[tdf.year_col])
    
    # Determine trend month (shifted from Calender months)
    # Adjustements by 1 are for converting between 0- and 1-based index
    df[tdf.trend_month_col] = ((df[tdf.month_col] + (12-trend_end_month) - 1)
                                           %12 + 1)
    
    # Determine trend season (shifted from Calendar seasons)
    df[tdf.trend_season_col] = ((df[tdf.trend_month_col] - 1)
                                            //int(12/seasons_per_year) + 1)
    
    return df


def _season_reduction(tdf,
                      df,
                      trend_end_month,
                      groupby_cols=None,
                      reduction_method=None,
                      annual_midpoint_date=None):
    
    # Add created columns to groupby_cols
    if groupby_cols == None:
        reduction_groupby_cols = []
    else:
        reduction_groupby_cols = groupby_cols.copy()
    
    reduction_groupby_cols += [tdf.freq_col,
                               tdf.freq_num_col,
                               tdf.trend_year_col,
                               tdf.trend_season_col]
    
    # Check if there are any seasons with multiple results
    if not df.duplicated(subset = reduction_groupby_cols).any():
        return df
    # If multiple results, then check that reduction method is not None
    elif reduction_method == None:
        raise ValueError('Multiple results exist within at least one season. '
            'Resolve any instances or choose a reduction method so that there '
            'is only ever a single result within each season for each group.')
    # Else check for valid reduction method
    elif reduction_method not in ['first','last','midpoint',
                            'median', 'mean','average','maximum','minimum']:
        raise ValueError('The value provided for reduction_method is not '
            f'a valid option. The provided value was {reduction_method}')
    
    # Apply reduction method
    
    # If statistical reduction
    if reduction_method in ['median','mean','average','maximum','minimum']:
        # Check for censored data
        if tdf.censored_values:
            raise ValueError(f'The {reduction_method} reduction method is not '
                'available for censored data at this time. Choose another '
                'reduction method or reduce the data without this package.')
        # Apply simple stat reduction
        if reduction_method == 'median':
            df = df.groupby(reduction_groupby_cols)[tdf.value_col].median()
        elif reduction_method in ['mean','average']:
            df = df.groupby(reduction_groupby_cols)[tdf.value_col].mean()
        elif reduction_method == 'maximum':
            df = df.groupby(reduction_groupby_cols)[tdf.value_col].max()
        elif reduction_method == 'minimum':
            df = df.groupby(reduction_groupby_cols)[tdf.value_col].min()
        # Reset index
        df = df.reset_index()
    # If date based reduction, check for multiple values for same date or
    # for same month depending on supplied time format
    else:
        if tdf.date_col != None:
            if df.duplicated(subset = reduction_groupby_cols +
                                                     [tdf.date_col]).any():
                raise Exception('There are multiple values supplied for the '
                    'same DateTime within a group. Date based reduction '
                    'cannot be used as these values are not able to be '
                    'sorted. Resolve these values or choose a statistical '
                    'reduction method (min, median, mean, max).')
        else:
            if df.duplicated(subset = reduction_groupby_cols +
                                                     [tdf.month_col]).any():
                raise Exception('There are multiple values supplied for the '
                    'same month within a group. Date based reduction '
                    'cannot be used as these values are not able to be '
                    'sorted. Resolve these values or choose a statistical '
                    'reduction method (min, median, mean, max).')
    
    if reduction_method in ['first','last']:
        if tdf.date_col != None:
            # Sort by date column
            df = df.sort_values(by=tdf.date_col)
        else:
            # Sort by date column
            df = df.sort_values(by=tdf.month_col)
            
        # Keep first/last result in each season
        df = df.drop_duplicates(subset=reduction_groupby_cols,
                                keep=reduction_method)
    elif reduction_method == 'midpoint':
        # If date column provided, determine the date of the midpoint
        if tdf.date_col != None:
            
            # Only quarterly and monthly seasons have midpoints mid-month
            # Use midnight of the 16th (end of 15th)
            df[tdf._midpointday_col] = np.where(df[tdf.freq_num_col].isin([4,12]),
                                             16,
                                             1)
            
            # Determine the trend month that corresponds with the provided
            # annual midpoint month
            if annual_midpoint_date != None:
                # Correct midpoint day if there is an annual override date
                df[tdf._midpointday_col] = np.where(df[tdf.freq_num_col] == 1,
                                                 annual_midpoint_date['day'],
                                                 df[tdf._midpointday_col])
                # Determine the trend month that corresponds with the provided
                # annual midpoint month
                trend_annual_midpoint_month = (
                    (annual_midpoint_date['month'] - trend_end_month -1)%12 + 1
                    )
            else:
                trend_annual_midpoint_month = 7
            
            # Determine the different between 
            trend_annual_midpoint_month_diff = trend_annual_midpoint_month - 7
            
            
            # Determine the midpoint year and month for all seasons per year
            conditions = [
                # Use the same year and month of date for monthly data
                (df[tdf.freq_num_col] == 12),
                
                # For odd trend months (in first month of 2-month interval)
                # then use the year/month from the month that follows
                # Use the same year and month of date for bimonthly data
                # if the trend month is even (second month of 2-month period)
                # else add a month to the date and use that year/month
                (df[tdf.freq_num_col] == 6) & (df[tdf.trend_month_col]%2 == 1),
                (df[tdf.freq_num_col] == 6) & (df[tdf.trend_month_col]%2 == 0),
                
                # Use similar logic for quarterly data
                (df[tdf.freq_num_col] == 4) & (df[tdf.trend_month_col]%3 == 1),
                (df[tdf.freq_num_col] == 4) & (df[tdf.trend_month_col]%3 == 2),
                (df[tdf.freq_num_col] == 4) & (df[tdf.trend_month_col]%3 == 0),
                
                # Use similar logic for triannual data
                (df[tdf.freq_num_col] == 3) & (df[tdf.trend_month_col]%4 == 1),
                (df[tdf.freq_num_col] == 3) & (df[tdf.trend_month_col]%4 == 2),
                (df[tdf.freq_num_col] == 3) & (df[tdf.trend_month_col]%4 == 3),
                (df[tdf.freq_num_col] == 3) & (df[tdf.trend_month_col]%4 == 0),
                
                # Use similar logic for semi-annual data
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 1),
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 2),
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 3),
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 4),
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 5),
                (df[tdf.freq_num_col] == 2) & (df[tdf.trend_month_col]%6 == 0),
                
               # Use similar logic for annual data
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 1),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 2),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 3),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 4),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 5),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 6),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 7),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 8),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 9),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 ==10),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 ==11),
               (df[tdf.freq_num_col] == 1) & (df[tdf.trend_month_col]%12 == 0)
                ]
            
            midpoint_month = [
                df[tdf.month_col],
                
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.month,
                df[tdf.month_col],
                
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.month,
                df[tdf.month_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.month,
                
                (df[tdf.date_col] + pd.DateOffset(months=2)).dt.month,
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.month,
                df[tdf.month_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.month,
                
                (df[tdf.date_col] + pd.DateOffset(months=3)).dt.month,
                (df[tdf.date_col] + pd.DateOffset(months=2)).dt.month,
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.month,
                df[tdf.month_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.month,
                (df[tdf.date_col] + pd.DateOffset(months=-2)).dt.month,
                
                (df[tdf.date_col] + pd.DateOffset(
                    months=6 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=5 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=4 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=3 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=2 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=1 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=0 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-1 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-2 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-3 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-4 + trend_annual_midpoint_month_diff)
                    ).dt.month,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-5 + trend_annual_midpoint_month_diff)
                    ).dt.month
                
                ]
            
            midpoint_year = [
                df[tdf.year_col],
                
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.year,
                df[tdf.year_col],
                
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.year,
                df[tdf.year_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.year,
                
                (df[tdf.date_col] + pd.DateOffset(months=2)).dt.year,
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.year,
                df[tdf.year_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.year,
                
                (df[tdf.date_col] + pd.DateOffset(months=3)).dt.year,
                (df[tdf.date_col] + pd.DateOffset(months=2)).dt.year,
                (df[tdf.date_col] + pd.DateOffset(months=1)).dt.year,
                df[tdf.year_col],
                (df[tdf.date_col] + pd.DateOffset(months=-1)).dt.year,
                (df[tdf.date_col] + pd.DateOffset(months=-2)).dt.year,
                
                (df[tdf.date_col] + pd.DateOffset(
                    months=6 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=5 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=4 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=3 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=2 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=1 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=0 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-1 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-2 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-3 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-4 + trend_annual_midpoint_month_diff)
                    ).dt.year,
                (df[tdf.date_col] + pd.DateOffset(
                    months=-5 + trend_annual_midpoint_month_diff)
                    ).dt.year
                ]
            
        else:
            raise ValueError(f'The {reduction_method} reduction method is not '
                'available for year/month data at this time. Choose another '
                'reduction method or reduce the data to a single result per '
                'season without this package.')
            
        # Use conditions to get month and year for midpoint
        df[tdf._midpointmonth_col] = np.select(conditions,
                                           midpoint_month,
                                           np.nan)
        df[tdf._midpointyear_col] = np.select(conditions,
                                           midpoint_year,
                                           np.nan)
        # Construct midpoint date
        df[tdf._midpointdate_col] = pd.to_datetime(
            dict(year=df[tdf._midpointyear_col],
                 month=df[tdf._midpointmonth_col],
                 day=df[tdf._midpointday_col]))
        # Determine proximity of result to midpoint
        df[tdf._midpointproximity_col] = abs(
            df[tdf.date_col] - df[tdf._midpointdate_col])
        
        # Sort by proximity and use earlier result if ties
        df = df.sort_values(by=[tdf._midpointproximity_col, tdf.date_col])
        # Keep single result closest to midpoint date
        df = df.drop_duplicates(subset=reduction_groupby_cols)
        # Drop temporary column
        df = df.drop(columns=[tdf._midpointday_col, tdf._midpointmonth_col,
                    tdf._midpointyear_col, tdf._midpointdate_col,
                    tdf._midpointproximity_col])
    
    return df

def _trend_periods(tdf,
                   df,
                   trend_end_month,
                   trend_lengths,
                   end_years):

    
    
    # Initiate output DataFrame
    output_df = []
    
    # Loop through end dates
    for end_year in end_years:
        
        # Copy data
        data = df.copy()
        
        # Only consider data up through the end year
        data = data[data[tdf.trend_year_col] <= end_year]
        
        # Determine DateTime for the end of the trend period
        end = (pd.to_datetime('{}-{}-1'.format(end_year, trend_end_month))
                   + pd.DateOffset(months=1)
                   + pd.Timedelta(seconds=-1))
        
        # Add a column for the end of the trend, if output format specified
        if tdf.output_format_trend_end:
            data[tdf.trend_end_col] = end.strftime(tdf.output_format_trend_end)
        
        # Sort trend lengths in descending order
        # This allows the use of the same data variable for each trend length
        trend_lengths.sort(reverse=True)
        
        # Loop through trend lengths for each end date
        for trend_length in trend_lengths:
            
            # Determine the start year (the difference between the first year
            # and last year in a trend period is 1 less than the trend length)
            start_year = end_year - (trend_length - 1)
            
            # Only keep years after and including the start year
            data = data[data[tdf.trend_year_col] >= start_year]
            
            # Add a column for the trend period, if output format specified
            if tdf.output_format_trend_period:
                # Determine DateTime for the start of the trend period
                start = (pd.to_datetime('{}-{}-1'.format(end_year,
                                                         trend_end_month))
                            + pd.DateOffset(months=1)
                            + pd.DateOffset(years=-trend_length))
                
                data[tdf.trend_period_col] = (
                    start.strftime(tdf.output_format_trend_period) + ' to '
                    + end.strftime(tdf.output_format_trend_period))
            
            # Add additional information columns
            data[tdf.trend_len_col] = trend_length
            
            # Append trend data to output
            output_df.append(data.copy())
        
    # Combine dataframes
    output_df = pd.concat(output_df).reset_index(drop=True)
    
    return output_df



'''
This module includes functions for analysing a single trend within a DataFrame

'''

import pandas as pd
import numpy as np
import censoredsummarystats as css
from scipy import stats

#%% Trend analysis for an individual dataset

def trend_analysis(df, tdf, seasons_per_year, seasonal_test=None):
    
    # Copy the DataFrame
    df = df.copy()
    
    # Initialise an empty output series
    output = []
    
    # Describe the dataset depending on whether results are censored or numeric
    if tdf.censored_values:
        # Create censored dataframe object
        cdf = css.CensoredData(df, tdf.value_col, **tdf.censored_kwargs)
        # Describe data
        output.append(describe_censored_data(tdf, cdf))
        # Convert censored data to have common censorship thresholds
        df = prep_censored_data(tdf, cdf)
    else:
        output.append(describe_data(tdf, df))
        
    # Describe the season counts of the dataset
    output.append(season_counts(tdf, df, seasons_per_year))
    
    # Replace value col with numeric col, if censored values
    if tdf.censored_values:
        df[tdf.value_col] = df[cdf.numeric_col]
    
    # Sort by date
    df = df.sort_values(by=[tdf.trend_year_col, tdf.trend_month_col])
    
    # Perform a Seasonality test
    kw_output = kruskal_wallis_test(df, tdf)
    seasonality = kw_output[tdf.kw_seasonality_col]
    output.append(kw_output)
    
    # Determine seasonal test from seasonality test if seasonal_test is None
    if seasonal_test == None:
        if seasonality == 'Seasonal':
            seasonal_test = True
        else:
            seasonal_test = False
    
    # Perform a trend direction analysis
    output.append(trend_direction(df, tdf, seasonal_test))
    
    # Perform a trend magnitude analysis
    output.append(trend_magnitude(df, tdf, seasons_per_year, seasonal_test))
    
    return pd.concat(output)

def describe_data(tdf, df):
    
    # Determine summary stats of the numeric data
    count = len(df)
    min_value = df[tdf.value_col].min()
    median_value = df[tdf.value_col].median()
    average_value = df[tdf.value_col].mean()
    max_value = df[tdf.value_col].max()
    
    return pd.Series([count,min_value,median_value,average_value,max_value],
                     index=[tdf.count_col,
                            tdf.minimum_col,
                            tdf.median_col,
                            tdf.average_col,
                            tdf.maximum_col])

def describe_censored_data(tdf, cdf):
    
    # Isolate dataframe
    df = cdf.data
    
    # Determine summary stats of the numeric data
    count = len(df)
    # Uncensored data stats
    non_censored = df[df[cdf.censor_col]==''][cdf.numeric_col]
    min_value = non_censored.min()
    max_value = non_censored.max()
    # Left censored data stats
    left_censored = df[df[cdf.censor_col].isin(['<','≤'])][cdf.numeric_col]
    count_left_censored = left_censored.count()
    min_detect_limit = left_censored.min()
    max_detect_limit = left_censored.max()
    # Right censored data stats
    right_censored = df[df[cdf.censor_col].isin(['≥','>'])][cdf.numeric_col]
    count_right_censored = right_censored.count()
    min_quant_limit = right_censored.min()
    max_quant_limit = right_censored.max()
    
    return pd.Series([count,min_value,max_value,
                      count_left_censored,min_detect_limit,max_detect_limit,
                      count_right_censored,min_quant_limit,max_quant_limit],
                     index=[tdf.count_col,
                            tdf.minimum_col,
                            tdf.maximum_col,
                            tdf.left_censored_count_col,
                            tdf.left_censored_min_col,
                            tdf.left_censored_max_col,
                            tdf.right_censored_count_col,
                            tdf.right_censored_min_col,
                            tdf.right_censored_max_col])

def prep_censored_data(tdf, cdf):
    
    # Isolate dataframe
    df = cdf.data
    # Create True/False indicator for uncensored data
    non_censored_check = (df[cdf.censor_col] == '')
    # Create True/False indicator for left censored data
    left_censored_check = df[cdf.censor_col].isin(['<','≤'])
    # Determine the maximum detection limit
    max_detect_limit = df[left_censored_check][cdf.numeric_col].max()
    # Create True/False indicator for right censored data
    right_censored_check = df[cdf.censor_col].isin(['≥','>'])
    # Determine the minimum quantification limit
    min_quant_limit = df[right_censored_check][cdf.numeric_col].min()
    
    # Convert all right censored data and all uncensored data that is greater
    # than the minimum quanitification limit to be equal to one
    # another and slightly larger than the limit
    if sum(right_censored_check) > 0:
        df[cdf.numeric_col] = np.where((right_censored_check) |
            (non_censored_check & (df[cdf.numeric_col] > min_quant_limit)),
            tdf.upper_conversion_factor*min_quant_limit,
            df[cdf.numeric_col])
    
    # Convert all left censored data and all uncensored data that is less
    # than the largest detection limit to be equal to one
    # another and less than the detection limit
    if sum(left_censored_check) > 0:
        df[cdf.numeric_col] = np.where((left_censored_check) | \
            (non_censored_check & (df[cdf.numeric_col] < max_detect_limit)),
            tdf.lower_conversion_factor*max_detect_limit,
            df[cdf.numeric_col])
    
    return df

def season_counts(tdf, df, seasons_per_year):
    
    # Determine trend_length and seasons per year
    trend_length = df[tdf.trend_len_col].iloc[0]
    
    years = len(df.groupby([tdf.trend_year_col]))
    seasons = len(df.groupby([tdf.trend_year_col,tdf.trend_season_col]))
    percent_years = 100 * years / trend_length
    percent_seasons = 100 * seasons / (seasons_per_year * trend_length)
    
    return pd.Series([years,seasons,percent_years,percent_seasons],
                     index=[tdf.years_in_trend_col,
                            tdf.seasons_in_trend_col,
                            tdf.percent_of_years_col,
                            tdf.percent_of_seasons_col])

#%% Seasonality Test

def kruskal_wallis_test(df, tdf):
    
    # Set default outputs
    KW_p = np.nan
    KW_seasonality = 'N/A'
        
    # Determine the unique seasons
    seasons = df[tdf.trend_season_col].unique()
    
    # Use default values if there are not multiple seasons
    if len(seasons) < 2:
        pass
    # If all values are identical, consider the result non-seasonal
    elif len(df[tdf.value_col].unique()) < 2:
        KW_p = 1.0
        KW_seasonality = 'Non-seasonal'
    else:
        # Perform the Kruskal-Wallis test
        KW = stats.kruskal(*[df[df[tdf.trend_season_col]==i][tdf.value_col]
                                                         for i in seasons])
        # Record the p-value
        KW_p = KW.pvalue
        # Using 0.05 as the cutoff, determine seasonality
        if KW_p <= tdf.seasonality_alpha:
            KW_seasonality = 'Seasonal'
        elif KW_p > tdf.seasonality_alpha:
            KW_seasonality = 'Non-seasonal'
    
    return pd.Series([KW_p,KW_seasonality],
                     index=[tdf.kw_pvalue_col,tdf.kw_seasonality_col])

#%% Trend Direction

def mann_kendall(df, tdf):
    
    # Convert Series to numpy array
    data = df[tdf.value_col].to_numpy()
    
    # Determine length of dataset
    n = data.size
    
    # Create array/matrix of increases (+1), decreases (-1), and ties (0)
    # to determine the Mann-Kendall S-statistic
    s = np.sign(data - data[:, None])
    # Sum the upper triangle of the matrix
    s = np.triu(s).sum()
    
    # Determine the unique values and number of multiple occurences(ties)
    counts = np.unique(data, return_counts=True)[1]
    
    # Use the number of repeated values to adjust the variance calculation
    if len(counts) == n:
        tt = 0
    else:
        counts = counts[counts>1]
        tt = np.sum(counts * (counts-1) * (2 * counts + 5))
    var = (n*(n-1)*(2*n+5)-tt)/18.
    
    return pd.Series([s,var],
                     index=[tdf.mk_svalue_col,tdf.mk_variance_col])

def mann_kendall_seasonal(df, tdf):
    
    # Group by season and sum each seasons S and variance to get
    # an overall S and variance
    output = (df.groupby(tdf.trend_season_col)
                  .apply(mann_kendall, tdf)
                  .sum())
    
    return output

def trend_direction(df, tdf, seasonal_test):
    
    # Determine whether to use seasonal or non-seasonal test based on
    # whether a season_col is provided.
    if seasonal_test:
        method = 'Seasonal'
        s, var = mann_kendall_seasonal(df, tdf)
    else:
        method = 'Non-seasonal'
        s, var = mann_kendall(df, tdf)
    
    # Calculate the Z-score using the S-statistic and the variance
    z = s - np.sign(s)
    if z != 0:
        z = z / np.sqrt(var)
    
    # Calculate the p-value from the Z-score
    # where the p-value is determined for a two-tailed test
    p = stats.norm.cdf(z)
    p = 2*min(p, 1-p)
    
    # Convert p-value to confidence in the trend direction
    C = 1 - 0.5*p
    
    # Convert confidence to a continuous scale of confidence that the trend is increasing
    if s < 0:
        C = 1 - C
    
    # Set default trend category
    trend = tdf.neutral_category
    
    # Sort confidence categories from largest to smallest
    confidence_categories = dict(sorted(tdf.confidence_categories.items(),
                                        reverse=True))
    
    # Convert confidence to a trend category
    for cutoff, category in confidence_categories.items():
        # Check cutoff values
        if cutoff >= 1.0 or cutoff <= 0.5:
            raise ValueError(f'A cutoff value of {cutoff} was included. '
                             'Cutoffs must be between 0.5 and 1.0.')
        if max(C, 1-C) >= cutoff:
            trend = category
            if C > 0.5:
                trend += ' increasing'
            else:
                trend += ' decreasing'
            break
    
    return pd.Series([method, s, var, p, 100*C, trend],
                     index=[tdf.applied_seasonality_col,
                            tdf.mk_svalue_col,
                            tdf.mk_variance_col,
                            tdf.mk_pvalue_col,
                            tdf.confidence_col,
                            tdf.trend_category_col])

#%% Trend Magnitude

def _slopes(df, tdf, seasons_per_year):
    
    # Copy the DataFrame
    df = df.copy()
    
    # Determine the decimal value of the year for use in slopes
    df['__DecimalYear__'] = df[tdf.trend_year_col]
    # Add partial year based on season
    df['__DecimalYear__'] += df[tdf.trend_season_col] / (seasons_per_year)
    
    # Set y values and x values
    y = np.array(df[tdf.value_col])
    x = np.array(df['__DecimalYear__'])

    # Compute sorted slopes only when deltax > 0
    deltax = x[:, np.newaxis] - x
    deltay = y[:, np.newaxis] - y
    slopes = deltay[deltax > 0] / deltax[deltax > 0]
    
    return slopes

def slopes_seasonal(df, tdf):
    
    # Determine the unique seasons
    seasons = df[tdf.trend_season_col].unique()
    
    slopes_seasonal = np.concatenate([
        _slopes(df[df[tdf.trend_season_col]==season], tdf, seasons_per_year=1)
                                                        for season in seasons])
    
    return slopes_seasonal


def trend_magnitude(df,
                    tdf,
                    seasons_per_year,
                    seasonal_test):
    
    # Determine whether to use seasonal or non-seasonal set of slopes based on
    # whether a season_col is provided.
    if seasonal_test:
        slopes = slopes_seasonal(df, tdf)
    else:
        slopes = _slopes(df, tdf, seasons_per_year)
    
    # Calculate median slope
    median_slope = np.nan
    if len(slopes) != 0:
        median_slope = np.median(slopes)
    
    # Calculate confidence interval
    
    # Initialise bounds of confidence interval
    lower_slope = np.nan
    upper_slope = np.nan
    
    # Calculate the percentiles of the confidence interval
    percentile_lower = (100 - tdf.confidence_interval)/2
    percentile_upper = 100 - percentile_lower
        
    # Map MfE method names to numpy method names
    mfe_to_numpy_percentile = {
        'weiball':'weibull',
        'tukey':'median_unbiased',
        'blom':'normal_unbiased',
        'hazen':'hazen',
        'excel':'linear'
        }
    
    if len(slopes) != 0:
        # Calculate percentiles for lower/upper bound of confidence interval
        lower_slope = np.percentile(slopes, percentile_lower,
                        method=mfe_to_numpy_percentile[tdf.percentile_method])
        
        upper_slope = np.percentile(slopes, percentile_upper,
                        method=mfe_to_numpy_percentile[tdf.percentile_method])        
    
    return pd.Series([median_slope,lower_slope,upper_slope],
                     index=[tdf.median_slope_col,
                            tdf.lower_slope_col,
                            tdf.upper_slope_col])


'''
These functions validate user input values.

'''

import pandas as pd

#%% Validate TrendData

def _validate_tdf(tdf):
    
    #%% Check data input
    
    # Check that supplied data is a DataFrame
    if not isinstance(tdf.data, pd.core.frame.DataFrame):
        raise ValueError('The data supplied to TrendData() must be a '
            'pandas DataFrame. Instead, an object was passed with type: '
            f'{type(tdf.data).__name__}')
    
    # Ensure that provided data is a copy and not indexed
    tdf.data = tdf.data.copy().reset_index()
    
    #%% Check value_col exists as column in data
    
    if tdf.value_col not in tdf.data.columns:
        raise ValueError('The value column supplied to TrendData() '
            'was not found as a column in the data. Columns include '
            f'{tdf.data.columns.to_list()} which does not include '
            f'{repr(tdf.value_col)}.')
    
    #%% Check that there are no nan values or empty strings
    if ((tdf.data[tdf.value_col].isnull().any()) |
        (tdf.data[tdf.value_col].astype(str).str.len().min() == 0)):
        raise ValueError('Missing values need to be removed from the data '
            'before it can be analysed.')
    
    #%% Check year, month, date columns
    if (((tdf.date_col == None) &
        ((tdf.year_col == None) | (tdf.month_col == None))) |
        ((tdf.date_col != None) &
        ((tdf.year_col != None) | (tdf.month_col != None)))):
        raise ValueError('Either specify date_col or specify both year_col '
            'and month_col.')
    
    # Check that date column is datetime format
    if tdf.date_col != None:
        if not (pd.api.types.is_datetime64_any_dtype(
                                    tdf.data[tdf.date_col].dtype)):
            raise ValueError('The values provided in the column '
                f'{tdf.date_col} must be datetime data type. Try using '
                'pd.to_datetime() to convert values.')
    
    # Check that year month columns are integer and month between 1 and 12
    if tdf.date_col == None:
        if not pd.api.types.is_integer_dtype(tdf.data[tdf.year_col]):
            raise ValueError('The values provided in the column '
                f'{tdf.year_col} must be integers.')
        if not pd.api.types.is_integer_dtype(tdf.data[tdf.month_col]):
            raise ValueError('The values provided in the column '
                f'{tdf.month_col} must be integers.')
        if ((tdf.data[tdf.month_col].max() > 12) |
            (tdf.data[tdf.month_col].min() < 1)):
            raise ValueError('The values provided in the column '
                f'{tdf.month_col} must be between 1 and 12.')
    
    #%% Check the input settings
    
    setting_mode = getattr(tdf, 'censored_values')
    if not isinstance(setting_mode, bool):
        raise ValueError(f'The value supplied to censored_values must be '
            'True or False. Instead, the value supplied was '
            f'{repr(setting_mode)}.')
        
    
    #%% Check column names for conflicts
    
    # Create list of column names that may be created
    created_columns = [attribute for attribute in tdf.__annotations__
                       if attribute.endswith('_col') and
                           attribute not in ['value_col',
                                             'year_col',
                                             'month_col',
                                             'date_col']]
    
    # Check the list for duplicates
    if len(created_columns) != len(set(created_columns)):
        raise ValueError('Two default column names have been set to the '
            'same name. Check that assigned values are unique and not '
            'identical to any default values.')
    
    
    for col in created_columns:
        col_name = getattr(tdf, col)
        # Check that columns created within methods don't conflict with column
        # names in provided data
        if (col_name in tdf.data.columns) & (col_name != tdf.value_col):
            raise ValueError('The data contains a column named '
                f'"{col_name}" which may be used in the output. Either '
                'rename this column or provide an alternative value for '
                f'{col} when initialising TrendData object.')
        # Check that new column names are string values
        if not isinstance(col_name, str):
            raise ValueError('Text values are required for column names. '
                f'The column name {col_name} should be changed to a text '
                f'value for {col}.')
    
    # Check that the data has no columns starting with double underscore
    dunder_cols = [col for col in tdf.data if col.startswith('__')]
    if len(dunder_cols) != 0:
        raise ValueError('Columns starting with "__" are used within '
            'the stat methods. Rename the following columns: '
            f'{dunder_cols}')
    
    #%% Check conversion factors
    if tdf.censored_values:
        # Check that the upper conversion factor is greater than 1
        if tdf.upper_conversion_factor <= 1:
            raise ValueError('The upper conversion factor needs to be greater '
                f'than 1. {tdf.upper_conversion_factor} was passed.')
        # Check that the lower conversion factor is less than 1
        if tdf.lower_conversion_factor >= 1:
            raise ValueError('The lower conversion factor needs to be less '
                f'than 1. {tdf.lower_conversion_factor} was passed.')

#%% Validate trend_end_month
def _validate_trend_end_month(end_month):
    
    # Check that end month is an integer
    if not isinstance(end_month, int):
        raise ValueError('The value supplied to trend_end_month must be '
            f'an integer. The value supplied was {end_month}')
    
    # Check that the end month provided is between 1 and 12
    if end_month - 1 not in range(12):
        raise ValueError('The value for trend_end_month must be between 1 and '
            f'12. The value provided was {end_month}')
        
#%% Validate seasons_per_year
def _validate_seasons_per_year(seasons_per_year):
    
    # Check that seasons_per_year is an integer
    if not isinstance(seasons_per_year, int):
        raise ValueError('The value supplied to seasons_per_year must be '
            f'an integer. The value supplied was {seasons_per_year}')
    
    # Check that the end month provided is between 1 and 12
    if seasons_per_year not in [1,2,3,4,6,12]:
        raise ValueError('The value for seasons_per_year must be a factor of '
            f'12. The value provided was {seasons_per_year}')

#%% Validate groupby columns

def _validate_groupby_cols(tdf, groupby_cols):
    
    # Skip None case
    if groupby_cols == None:
        pass
    else:
        # Check that groupby_cols is a list
        if not isinstance(groupby_cols, list):
            raise ValueError('groupby_cols needs to be a list. '
                f'The provided input was {groupby_cols}')
        
        # Check that column names are included in data
        if not (all(name in tdf.data.columns for name in groupby_cols)):
            raise ValueError('Columns used in groupby_cols were not found in '
                'the DataFrame. Names not found include: '
                f'{[x for x in groupby_cols if x not in tdf.data.columns]}')
        
        # Check for null values in groupby columns
        if tdf.data[groupby_cols].isnull().values.any():
            raise ValueError('Null values found in one of the columns used '
                'for grouping.')

#%% Validate annual midpoint date

def _validate_annual_midpoint_date(annual_midpoint_date):
    
    # Skip None case
    if annual_midpoint_date == None:
        pass
    else:
        # Check that annual_midpoint_date is a dictionary
        if not isinstance(annual_midpoint_date, dict):
            raise ValueError('annual_midpoint_date needs to be a dictionary. '
                f'The provided input was {annual_midpoint_date}')
        
        # Check that key names are only month and day
        if list(annual_midpoint_date.keys()) != ['month','day']:
            raise ValueError('annual_midpoint_date requires a dictionary with '
                'month and day provided as keys. The provided keys were '
                f'{list(annual_midpoint_date.keys())}')
        
        # Check that month value is integer between 1 and 12
        if ((not isinstance(annual_midpoint_date['month'], int)) |
                (annual_midpoint_date['month']-1 not in range(12))):
            raise ValueError('The value supplied as the month in '
                'annual_midpoint_date must be an integer between 1 and 12.'
                f'The value supplied was {annual_midpoint_date["month"]}')
        
        # Check that day value is integer between 1 and 31
        if ((not isinstance(annual_midpoint_date['day'], int)) |
                (annual_midpoint_date['day']-1 not in range(31))):
            raise ValueError('The value supplied as the month in '
                'annual_midpoint_date must be an integer between 1 and 31.'
                f'The value supplied was {annual_midpoint_date["day"]}')

#%% Validate trend lengths and end years

def _validate_trend_lengths_and_end_years(trend_lengths, end_years):
    
    # Check that trend lengths input is a list
    if not isinstance(trend_lengths, list):
        raise ValueError('trend_lengths needs to be a list. '
            f'The provided input was {trend_lengths}')
    # Check that the provided values are integers
    if not (all(isinstance(length, int) for length in trend_lengths)):
        raise ValueError('All trend lengths need to be integers. '
            f'The provided trend lengths were {trend_lengths}')
    # Check that trend_lengths are positive
    if not (all(length > 0 for length in trend_lengths)):
        raise ValueError('All trend lengths need to be positive. '
            f'The provided trend lengths were {trend_lengths}')
    
    # Check that end years input is a list
    if not isinstance(end_years, list):
        raise ValueError('end_years needs to be a list. '
            f'The provided input was {end_years}')
    # Check that the provided values are integers
    if not (all(isinstance(year, int) for year in end_years)):
        raise ValueError('All end years need to be integers. '
            f'The provided end years were {end_years}') 


# -*- coding: utf-8 -*-
"""
Created on Sun Jun 25 11:28:27 2023

@author: KurtV
"""

import pandas as pd
from dataclasses import dataclass, field
from environmentaltrends.validation import (
    _validate_tdf,
    _validate_trend_end_month,
    _validate_seasons_per_year,
    _validate_groupby_cols,
    _validate_annual_midpoint_date,
    _validate_trend_lengths_and_end_years)
from environmentaltrends.data_prep import (
    _define_seasons,
    _season_reduction,
    _trend_periods)
from environmentaltrends.single_trend_analysis import trend_analysis


@dataclass
class TrendData:
    '''
    This dataclass validates and pre-processes a dataframe for generating
    trends.

    Parameters
    ----------
    data : DataFrame
        A pandas dataframe containing time series data
    value_col : string
        The column name for the column with censored results
    date_col : string
        The column name for the column containing DateTime information
    year_col : string
        Alternative option for a DateTime column. The column name for the
        column containing the year.
    month_col : string
        Alternative option for a DateTime column. The column name for the
        column containing the month as an integer.
    seasonality_alpha : float
        The p-value threshold for determining seasonality using the Kruskall-
        Wallis test. The default is 0.05.
    confidence_categories : dictionary of float keys and string values
        The cutoffs used to establish trend likelihood categories. The default
        is {0.90: 'Very likely', 0.67: 'Likely'}, meaning 90%+ confidence is
        considered 'very likely' and 67%+ is consider 'likely'.
    neutral_category : string
        The category name for trends that do not meet the minimum threshold
        set in confidence_categories. The default is 'Indeterminate'.
    confidence_interval : float
        The confidence interval to use for determining the upper and lower
        slope estimates. The default is 90.
    percentile_method : string
        The percentile method to use for setting the confidence interval for
        slopes. The default is 'hazen'. Additional options include 'weiball',
        'tukey', 'blom', 'excel'.
    output_format_trend_end : string
        The format to use in the output column containing the end of the
        trend period. The default is '%Y' which produces the year only.
    output_format_trend_period : string
        The format to use in the output column containing the entire trend
        period. The default is '%b %Y' to output month and year for the start
        and end of the trend period.
    censored_values: boolean
        An indicator for whether censored values are expected in the value_col.
        The default is False.
    censored_kwargs : dictionary
        A dictionary of keywords and values to use for generation of a
        CensoredData object. Users should be familiar with the Python package
        censoredsummarystats.
    lower_conversion_factor : float
        The conversion factor to use for left censored values.
        The default is 0.5 which assumes positive values only.
    upper_conversion_factor : float
        The conversion factor to use for right censored values.
        The default is 1.1 which assumes positive values only.
    '''
    data: pd.core.frame.DataFrame
    value_col: any
    date_col: str = None
    year_col: str = None
    month_col: str = None
    seasonality_alpha: float = 0.05
    confidence_categories: dict[float,str] = field(
        default_factory=lambda: {0.90: 'Very likely', 0.67: 'Likely'}
    )
    neutral_category: str = 'Indeterminate'
    confidence_interval: float = 90
    percentile_method: str = 'hazen'
    output_format_trend_end: str = '%Y'
    output_format_trend_period: str = '%b %Y'
    censored_values: bool = False
    censored_kwargs: dict = field(default_factory=dict)
    lower_conversion_factor: float = 0.5
    upper_conversion_factor: float = 1.1
    freq_col: str = 'Frequency'
    freq_num_col: str = 'SeasonsPeryear'
    trend_year_col:str = 'TrendYear'
    trend_month_col:str = 'TrendMonth'
    trend_season_col:str = 'TrendSeason'
    trend_end_col: str = 'TrendEnd'
    trend_len_col: str = 'TrendLength'
    trend_period_col: str = 'TrendPeriod'
    count_col: str = 'ValueCount'
    minimum_col: str = 'Minimum'
    median_col: str = 'Median'
    average_col: str = 'Average'
    maximum_col: str = 'Maximum'
    left_censored_count_col: str = 'CountBelowDetect'
    left_censored_min_col: str = 'DetectLimitMin'
    left_censored_max_col: str = 'DetectLimitMax'
    right_censored_count_col: str = 'CountAboveQuant'
    right_censored_min_col: str = 'QuantLimitMin'
    right_censored_max_col: str = 'QuantLimitMax'
    years_in_trend_col: str = 'YearsInPeriod'
    seasons_in_trend_col: str = 'SeasonsInPeriod'
    percent_of_years_col: str = 'PercentOfYears'
    percent_of_seasons_col: str = 'PercentOfSeasons'
    kw_pvalue_col: str = 'KW-pValue'
    kw_seasonality_col: str = 'Seasonality'
    applied_seasonality_col: str = 'AppliedSeasonality'
    mk_svalue_col: str = 'MK-S'
    mk_variance_col: str = 'MK-Variance'
    mk_pvalue_col: str = 'MK-pvalue'
    confidence_col: str = 'IncreasingLikelihood'
    trend_category_col: str = 'TrendDirection'
    median_slope_col: str = 'SenSlope'
    lower_slope_col: str = 'LowerSlope'
    upper_slope_col: str = 'UpperSlope'
    _midpointday_col: str = field(default='__tempMidPointDay__', repr=False)
    _midpointmonth_col: str = field(
        default='__tempMidPointMonth__', repr=False)
    _midpointyear_col: str = field(default='__tempMidPointYear__', repr=False)
    _midpointdate_col: str = field(default='__tempMidPointDate__', repr=False)
    _midpointproximity_col: str = field(
        default='__tempMidPointProximity__', repr=False)
    
    
    def __post_init__(self):
        
        #%% Validate inputs
        
        _validate_tdf(self)
        
        #%% Convert values to float if censored_values is False
        if not self.censored_values:
            try:
                self.data[self.value_col] = (self.data[self.value_col]
                                                 .astype(float))
            except:
                raise ValueError('Values could not be converted to a '
                    'float data type. If values contain censor characters '
                    'such as < or > then set censored_values = True')
        
        #%% Complete year and month info
        if self.date_col != None:
            self.year_col = '__tempCalendarYear__'
            self.month_col = '__tempCalendarMonth__'
            self.data[self.year_col] = self.data[self.date_col].dt.year
            self.data[self.month_col] = self.data[self.date_col].dt.month
        
    #%% Trend methods
    
    def trends(self,
               seasons_per_year,
               trend_lengths,
               end_years,
               groupby_cols=None,
               seasonal_test=None,
               trend_end_month=6,
               reduction_method=None,
               annual_midpoint_date=None):
        '''
        This method outputs a DataFrame containing trend results which include
        likelihood of trend direction, trend slope, and a confidence interval
        for the trend slope.

        Parameters
        ----------
        seasons_per_year : int
            The number of seasons that should be analysed within a given year.
            Accepted values are 1,2,3,4,6,12, where a value of 1 would generate
            annual trends and 12 would generate monthly trends.
        trend_lengths : list of integers
            A list of the desired trend lengths. Lists can include a single
            value or multiple values.
        end_years : list of integers
            A list of years that describe when the trend period ends.
        groupby_cols : list
            A list of columns that should be used to differentiate time series.
            The default is None.
        seasonal_test : boolean, optional
            Specify whether seasonal or non-seasonal trend analyses should be
            forced regardless of the seasonality test results. The default is
            None which will apply the trend analyses indicated by the
            seasonality test.
        trend_end_month : int, optional
            Trends don't have to align with the calendar year. This setting
            specifies the last month in the trend period.
            The default is 6 (June) to align with water years.
        reduction_method : string, optional
            When multiple results exist within a single season, trend analyses
            require that they are reduced to a single value. Options for this
            include using a statistical result ('median','mean','average',
            'maximum','minimum') or a temporal result ('first','last',
            'midpoint'), where midpoint is the result closest to the midpoint
            of the season. The default is None.
        annual_midpoint_date : dictionary with month and day keywords, optional
            If the reduction_method is 'midpoint' and seasons_per_year is 1,
            then an alternative target date can be used for reducing annual
            trend data. For example, to target values closest to the middle
            of New Zealand spring, use: {'month':10, 'day':16}.
            The default is None.
        
        Returns
        -------
        DataFrame
            Contains a DataFrame of trend results.

        '''
        
        # Validation
        _validate_trend_end_month(trend_end_month)
        _validate_seasons_per_year(seasons_per_year)
        _validate_groupby_cols(self, groupby_cols)
        _validate_annual_midpoint_date(annual_midpoint_date)
        _validate_trend_lengths_and_end_years(trend_lengths, end_years)
        
        # Classify results into a trend season based on season frequency and trend year
        df = _define_seasons(self, seasons_per_year, trend_end_month)
        
        # Create empty list if no groups
        if groupby_cols == None:
            groupby_cols = []
        
        # Reduce multiple results within a season to a single result
        if df.duplicated(subset = groupby_cols +
                                   [self.trend_year_col,
                                    self.trend_season_col]).any():
            df = _season_reduction(self, df, trend_end_month, groupby_cols,
                                   reduction_method, annual_midpoint_date)
            
        # Create dataset for each trend period
        df = _trend_periods(self,
                            df,
                            trend_end_month,
                            trend_lengths,
                            end_years)
        
        # Create list of columns to group by
        trend_groups = groupby_cols + [self.freq_col,
                                          self.freq_num_col,
                                          self.trend_len_col]
        if self.output_format_trend_end != None:
            trend_groups += [self.trend_end_col]
        if self.output_format_trend_period != None:
            trend_groups += [self.trend_period_col]
        
        # Analyse each dataset
        df = df.groupby(trend_groups).apply(trend_analysis,
                                            self,
                                            seasons_per_year,
                                            seasonal_test)
        
        # Reset index
        df = df.reset_index()
        
        return df